#!python3

import numpy as np
import oineus as oin
import dionysus as dion

def dion_dgm_to_numpy(dion_dgm, include_inf_points=True):
    return np.array([[p.birth, p.death] for p in dion_dgm if include_inf_points or np.abs(p.birth) < np.inf and np.abs(p.death) < np.inf]).astype(np.float32)

def helper_test_random(n, dualize, negate, n_threads, seed=1, top_dim=3, compute_v=False, compute_u=False):
    # no wrap in Dionysus
    wrap = False

    # generate random grid data
    np.random.seed(seed)
    a = np.random.randn(n ** 3).reshape((n, n, n))

    rp = oin.ReductionParams()
    rp.n_threads = n_threads
    rp.compute_v = compute_v
    rp.compute_u = compute_u
    # compute diagrams with Oineus
    oin_dgms = oin.compute_diagrams_ls(a, negate, wrap, top_dim-1, rp, include_inf_points=True, dualize=dualize)

    # compute diagrams with Dionysis
    fil_us = dion.fill_freudenthal(a, reverse=negate)
    p = dion.homology_persistence(fil_us)
    dion_dgms = dion.init_diagrams(p, fil_us)

    dist = 0.0

    for dim in range(top_dim):
        # convert Oineus diagram to Dionysus format
        oin_dgm = dion.Diagram(oin_dgms[dim])
        dion_dgm = dion_dgms[dim]
        dion_dgm = dion_dgm_to_numpy(dion_dgm, True)
        if negate:
            dion_dgm[dion_dgm == np.inf] = -np.inf
        dion_dgm = dion.Diagram(dion_dgm)
        dist += dion.bottleneck_distance(oin_dgm, dion_dgm)

    assert(dist < 0.001)


def test_random():
    for n in [4, 8, 17]:
        for dualize in [True, False]:
            for negate in [True, False]:
                for compute_u in [True, False]:
                    for compute_v in [True, False]:
                        for n_threads in [1, 4, 7]:
                            if n_threads > 1 and compute_u:
                                continue
                            helper_test_random(n=n, negate=negate, n_threads=n_threads, dualize=dualize, compute_v=compute_v, compute_u=compute_u)


if __name__ == "__main__":
    test_random()



#!/usr/bin/env python3

import argh
import argparse
import numpy as np

import oineus as oin
import dionysus as dion


def compare_diagrams(n=16, negate=False, n_threads=1, seed=1, top_dim=2):
    # no wrap in Dionysus
    wrap = False

    # generate random grid data
    np.random.seed(seed)
    a = np.random.randn(n ** 3).reshape((n, n, n))

    params = oin.ReductionParams()
    params.n_threads = n_threads

    # compute diagrams with Oineus
    oin_dgms = oin.compute_diagrams_ls(a, negate, wrap, top_dim, params, include_inf_points=True, dualize=False)

    # compute diagrams with Dionysis
    fil_us = dion.fill_freudenthal(a, reverse=negate)
    p = dion.homology_persistence(fil_us)
    dion_dgms = dion.init_diagrams(p, fil_us)

    dist = 0.0

    for dim in range(top_dim):
        # convert Oineus diagram to Dionysus format
        oin_dgm = dion.Diagram(oin_dgms[dim])
        dion_dgm = dion_dgms[dim]
        dist += dion.bottleneck_distance(oin_dgm, dion_dgm)

    print("total dist: ", dist)
    assert(dist < 0.001)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    argh.add_commands(parser, [compare_diagrams])
    argh.dispatch(parser)



from __future__ import absolute_import

import typing
import numpy as np
import scipy.sparse

from . import _oineus

from ._oineus import ConflictStrategy, DenoiseStrategy, VREdge
from ._oineus import CombinatorialProdSimplex, CombinatorialSimplex,Simplex, ProdSimplex
from ._oineus import Filtration, ProdFiltration
from ._oineus import Decomposition, IndexDiagramPoint, DiagramPoint, Diagrams
from ._oineus import ReductionParams, KICRParams, KerImCokReduced, KerImCokReducedProd
from ._oineus import IndicesValues, IndicesValuesProd, TopologyOptimizer, TopologyOptimizerProd
from ._oineus import compute_relative_diagrams, get_boundary_matrix, get_denoise_target, get_induced_matching
from ._oineus import get_nth_persistence, get_permutation_dtv, list_to_filtration

# import warnings
#
# try:
#     from . import diff
# except:
#     warnings.warn("oineus.diff import failed, probably, because eagerpy is not installed")
#

__all__ = ["compute_diagrams_ls", "compute_diagrams_vr", "get_boundary_matrix", "is_reduced"]


def to_scipy_matrix(sparse_cols, shape=None):
    if shape is None:
        shape = (len(sparse_cols), len(sparse_cols))
    row_ind = [j for i in range(len(sparse_cols)) for j in sparse_cols[i]]
    col_ind = [i for i in range(len(sparse_cols)) for _ in sparse_cols[i]]
    assert (len(row_ind) == len(col_ind))
    data = [1 for _ in range(len(row_ind))]
    return scipy.sparse.csc_matrix((data, (row_ind, col_ind)), shape=shape)


def max_distance(data: np.ndarray, from_pwdists: bool=False):
    if from_pwdists:
        return 1.00001 * np.min(np.max(data, axis=1))
    else:
        assert data.ndim == 2 and data.shape[0] >= 2
        diff = data[:, np.newaxis, :] - data[np.newaxis, :, :]
        squared_distances = np.sum(diff**2, axis=2)
        return 1.00001 * np.sqrt(np.min(np.max(squared_distances, axis=1)))


def freudenthal_filtration(data: np.ndarray,
                           negate: bool=False,
                           wrap: bool=False,
                           max_dim: int = 3,
                           with_critical_vertices: bool=False,
                           n_threads: int=1):
    max_dim = min(max_dim, data.ndim)
    if with_critical_vertices:
        fil, vertices = _oineus.get_freudenthal_filtration_and_crit_vertices(data=data, negate=negate, wrap=wrap, max_dim=max_dim, n_threads=n_threads)
        vertices = np.array(vertices, dtype=np.int64)
        return fil, vertices
    else:
        return _oineus.get_freudenthal_filtration(data=data, negate=negate, wrap=wrap, max_dim=max_dim, n_threads=n_threads)


def vr_filtration(data: np.ndarray,
                  from_pwdists: bool = False,
                  max_dim: int = -1,
                  max_diameter: float = -1.0,
                  with_critical_edges: bool = False,
                  n_threads: int = 1):
    assert data.ndim == 2

    if from_pwdists:
        assert data.shape[0] == data.shape[1]

    if max_diameter < 0:
        max_diameter = max_distance(data, from_pwdists)

    if max_dim < 0:
        if from_pwdists:
            raise RuntimeError("vr_filtration: if input is pairwise distance matrix, max_dim must be specified")
        else:
            max_dim = data.shape[1]

    if from_pwdists:
        if with_critical_edges:
            func = _oineus.get_vr_filtration_and_critical_edges_from_pwdists
        else:
            func = _oineus.get_vr_filtration_from_pwdists
    else:
        if with_critical_edges:
            func = _oineus.get_vr_filtration_and_critical_edges
        else:
            func = _oineus.get_vr_filtration

    result = func(data, max_dim=max_dim, max_diameter=max_diameter, n_threads=n_threads)
    if with_critical_edges:
        # convert list of VREdges to numpy array
        edges = [ [ e.x, e.y] for e in result[1] ]
        edges = np.array(edges, dtype=np.int64)
        return result[0], edges
    else:
        return result


def is_reduced(a):
    lowest_ones = []
    for col_idx in range(a.shape[1]):
        if np.any(a[:, col_idx] % 2 == 1):
            lowest_ones.append(np.max(np.where(a[:, col_idx] % 2 == 1)))
    return len(lowest_ones) == len(set(lowest_ones))

def mapping_cylinder(fil_domain, fil_codomain, v_domain, v_codomain, with_indices=False):
    if isinstance(v_codomain, _oineus.Simplex) or isinstance(v_codomain, _oineus.ProdSimplex):
        v_codomain = v_codomain.combinatorial_cell()
    if isinstance(v_domain, _oineus.Simplex) or isinstance(v_domain, _oineus.ProdSimplex):
        v_domain = v_domain.combinatorial_cell()
    if with_indices:
        return _oineus._mapping_cylinder_with_indices(fil_domain, fil_codomain, v_domain, v_codomain)
    else:
        return _oineus._mapping_cylinder(fil_domain, fil_codomain, v_domain, v_codomain)

def multiply_filtration(fil, sigma):
    if isinstance(sigma, _oineus.Simplex):
        sigma = sigma.combinatorial_cell()
    return _oineus._multiply_filtration(fil, sigma)

def min_filtration(fil_1, fil_2, with_indices=False):
    if with_indices:
        return _oineus._min_filtration_with_indices(fil_1, fil_2)
    else:
        return _oineus._min_filtration(fil_1, fil_2)

def compute_diagrams_ls(data: np.ndarray, negate: bool=False, wrap: bool=False,
                        max_dim: typing.Optional[int]=None, params: typing.Optional[ReductionParams]=None,
                        include_inf_points: bool=True, dualize: bool=False):
    if max_dim is None:
        max_dim = data.ndim - 1
    if params is None:
        params = _oineus.ReductionParams()
    # max_dim is maximal dimension of the _diagram_, we need simplices one dimension higher, hence +1
    fil = freudenthal_filtration(data=data, negate=negate, wrap=wrap, max_dim=max_dim + 1, n_threads=params.n_threads)
    dcmp = _oineus.Decomposition(fil, dualize)
    dcmp.reduce(params)
    return dcmp.diagram(fil=fil, include_inf_points=include_inf_points)


def compute_diagrams_vr(data: np.ndarray, from_pwdists: bool=False, max_dim: int=-1, max_diameter: float = -1.0, params: typing.Optional[ReductionParams]=None, include_inf_points: bool=True, dualize: bool=True):
    if params is None:
        params = _oineus.ReductionParams()
    # max_dim is maximal dimension of the _diagram_, we need simplices one dimension higher, hence +1
    fil = vr_filtration(data, from_pwdists, max_dim=max_dim, max_diameter=max_diameter, with_critical_edges=False, n_threads=params.n_threads)
    dcmp = _oineus.Decomposition(fil, dualize)
    dcmp.reduce(params)
    return dcmp.diagram(fil=fil, include_inf_points=include_inf_points)


def get_ls_wasserstein_matching_target_values(dgm, fil, rv, d: int, q: float, mip: bool, mdp: bool):
    func = getattr(_oineus, f"get_ls_wasserstein_matching_target_values")

    if type(dgm) is np.ndarray:
        dgm_1 = []
        assert len(dgm.shape) == 2 and dgm.shape[1] == 2
        for p in dgm:
            dgm_1.append(DiagramPoint(p[0], p[1]))
        dgm = dgm_1

    return func(dgm, fil, rv, d, q, mip, mdp)


def compute_kernel_image_cokernel_reduction(K, L, params=None, reduction_params=None):
    # simplicial filtrations can be supplied as lists,
    # convert to Oineus filtrations if necessary
    if isinstance(K, list):
        K = _oineus.list_to_filtration(K)
    if isinstance(L, list):
        L = _oineus.list_to_filtration(L)

    # KICR class is templatized by cell type in C++
    # different instantiations have different class names in Python
    # figure out the right one by type
    if isinstance(K[0], _oineus.Simplex):
        KICR_Class = _oineus.KerImCokReduced
    elif isinstance(K[0], _oineus.ProdSimplex):
        KICR_Class = _oineus.KerImCokReducedProd

    if params is None:
        # compute all by default
        params = _oineus.KICRParams()
        params.kernel = True
        params.cokernel = True
        params.image = True

    if reduction_params != None:
        assert type(params) == _oineus.ReductionParams
        params.params_f = reduction_params
        params.params_g = reduction_params
        params.params_ker = reduction_params
        params.params_im = reduction_params
        params.params_cok = reduction_params

    return KICR_Class(K, L, params)


def compute_ker_cok_reduction_cyl(fil_2, fil_3):
    fil_min = _oineus.min_filtration(fil_2, fil_3)

    id_domain = fil_3.size() + fil_min.size() + 1
    id_codomain = id_domain + 1

    # id_domain: id of vertex at the top of the cylinder,
    # i.e., we multiply fil_3 with id_domain
    # id_codomain: id of vertex at the bottom of the cylinder
    # i.e, we multiply fil_min with id_codomain

    v0 = _oineus.Simplex(id_domain, [id_domain])
    v1 = _oineus.Simplex(id_codomain, [id_codomain])

    fil_cyl = _oineus.mapping_cylinder(fil_3, fil_min, v0, v1)

    # to get a subcomplex, we multiply each fil_3 with id_domain
    fil_3_prod = _oineus.multiply_filtration(fil_3, v0)

    params = _oineus.KICRParams()
    params.kernel = params.cokernel = True
    params.image = False

    kicr_reduction = _oineus.KerImCokReducedProd(fil_cyl, fil_3_prod, params)

    return kicr_reduction


import oineus as oin
import math

def test_kernel_1():
	params=oin.ReductionParams()
	params.n_threads=4
	params.kernel=True
	params.image=True
	params.cokernel=True
	K = [ [0, [0], 10], [1,[1],50], [2,[2], 20], [3, [3], 50], [4,[4], 15], [5, [5], 12], [6,[0,1], 50], [7, [1,2], 60], [8,[2,3], 70], [9, [3,4], 80], [10, [0,5], 30], [11,[4,5], 20]]
	L = [ [0, [0], 10], [1,[1],50], [2,[2], 20], [3, [3], 50], [4,[4], 15], [5,[0,1], 50], [6, [1,2], 60], [7,[2,3], 70], [8, [3,4], 80] ]
	IdMapping = [0,1,2,3,4,6,7,8,9]
	kicr = oin.compute_kernel_image_cokernel_reduction(K, L, IdMapping, params)
	errors = False
	if (kicr.kernel_diagrams().in_dimension(0) != [[30., 80.],[20., math.inf]]).any():
		print("Error in kernel_test_1: kernel diagram in dimenion 0 incorrect.")
		errors = True
	if (kicr.kernel_diagrams().in_dimension(1) != []).any():
		print("Error in kernel_test_1: kernel diagram in dimenion 1 incorrect.")
		errors = True
	if (kicr.cokernel_diagrams().in_dimension(0) != [[12., 20.]]).any():
		print("Error in kernel_test_1: cokernel diagram in dimenion 0 incorrect.")
		errors = True
	if (kicr.cokernel_diagrams().in_dimension(1) != [[80., math.inf]]).any():
		print("Error in kernel_test_1: cokernel diagram in dimenion 1 incorrect.")
		errors = True
	if (kicr.image_diagrams().in_dimension(0) != [[15., 30.],[20., 60.],[50., 70.], [10., math.inf]]).any():
		print("Error in kernel_test_1: image diagram in dimenion 0 incorrect.")
		errors = True
	if (kicr.image_diagrams().in_dimension(1) != []).any():
		print("Error in kernel_test_1: image diagram in dimenion 0 incorrect.")
		errors = True
	if errors:
		return 1
	else:
		return 0
 

def test_kernel_2():
	params=oin.ReductionParams()
	params.n_threads=4
	params.kernel=True
	params.image=True
	params.cokernel=True
	K=[[0,[0], 10], [1, [1], 30], [2, [2], 10], [3, [3], 0], [4, [0,1], 30], [5, [1,2], 30], [6, [0,3], 10], [7, [2,3], 10]]
	L=[[0,[0], 10.], [1, [1], 30], [2, [2], 10], [3, [0,1], 30], [4, [1,2], 30]] 
	IdMapping=[0,1,2,4,5]
	kicr = oin.compute_kernel_image_cokernel_reduction(K, L, IdMapping, params)
	errors = False
	if (kicr.kernel_diagrams().in_dimension(0) != [[10., 30.]]).any():
		print("Error in kernel_test_1: kernel diagram in dimenion 0 incorrect.")
		errors = True
	if (kicr.kernel_diagrams().in_dimension(1) != []).any():
		print("Error in kernel_test_1: kernel diagram in dimenion 1 incorrect.")
		errors = True
	if (kicr.cokernel_diagrams().in_dimension(0) != [[0., 10.]]).any():
		print("Error in kernel_test_1: cokernel diagram in dimenion 0 incorrect.")
		errors = True
	if (kicr.cokernel_diagrams().in_dimension(1) != [[30., math.inf]]).any():
		print("Error in kernel_test_1: cokernel diagram in dimenion 1 incorrect.")
		errors = True
	if (kicr.image_diagrams().in_dimension(0) != [[10., math.inf]]).any():
		print("Error in kernel_test_1: image diagram in dimenion 0 incorrect.")
		errors = True
  	if (kicr.image_diagrams().in_dimension(1) != []).any():
		print("Error in kernel_test_1: image diagram in dimenion 0 incorrect.")
		errors = True
	if errors:
		return 1
	else:
		return 0




import numpy as np
import oineus as oin
f = np.random.randn(16, 16, 16)
params = oin.ReductionParams()
params.n_threads = 2
dgms = oin.compute_diagrams_ls(data=f, negate=False, wrap=False, params=params, include_inf_points=True, max_dim=2)
dgm = dgms.in_dimension(0)



import oineus as oin
import math

def test_kernel_1():
    params=oin.ReductionParams()
    params.n_threads=4
    params.kernel=True
    params.image=True
    params.cokernel=True
    K = [ [0, [0], 10], [1,[1],50], [2,[2], 20], [3, [3], 50], [4,[4], 15], [5, [5], 12], [6,[0,1], 50], [7, [1,2], 60], [8,[2,3], 70], [9, [3,4], 80], [10, [0,5], 30], [11,[4,5], 20]]
    L = [ [0, [0], 10], [1,[1],50], [2,[2], 20], [3, [3], 50], [4,[4], 15], [5,[0,1], 50], [6, [1,2], 60], [7,[2,3], 70], [8, [3,4], 80] ]
    IdMapping = [0,1,2,3,4,6,7,8,9]
    K = oin.list_to_filtration_double(K)
    L = oin.list_to_filtration_double(L)
    kicr = oin.KerImCokReduced_double(K, L, params)
    # print(kicr.kernel_diagrams().in_dimension(0))
    # assert (kicr.kernel_diagrams().in_dimension(0) == [[30., 80.],[20., math.inf]]).all()
    assert (len(kicr.kernel_diagrams().in_dimension(1)) == 0)
    assert (kicr.cokernel_diagrams().in_dimension(0) == [[12., 20.]]).all()
    assert (kicr.cokernel_diagrams().in_dimension(1) == [[80., math.inf]]).all()
    assert (kicr.image_diagrams().in_dimension(0) == [[15., 30.],[20., 60.],[50., 70.], [10., math.inf]]).all()
    assert (len(kicr.image_diagrams().in_dimension(1)) == 0)

def test_kernel_2():
    params=oin.ReductionParams()
    params.n_threads=4
    params.kernel=True
    params.image=True
    params.cokernel=True
    K=[[0,[0], 10], [1, [1], 30], [2, [2], 10], [3, [3], 0], [4, [0,1], 30], [5, [1,2], 30], [6, [0,3], 10], [7, [2,3], 10]]
    L=[[0,[0], 10.], [1, [1], 30], [2, [2], 10], [3, [0,1], 30], [4, [1,2], 30]]
    K = oin.list_to_filtration_double(K)
    L = oin.list_to_filtration_double(L)
    kicr = oin.KerImCokReduced_double(K, L, params)
    assert (kicr.kernel_diagrams().in_dimension(0) == [[10., 30.]]).all()
    assert (len(kicr.kernel_diagrams().in_dimension(1)) == 0)
    assert (kicr.cokernel_diagrams().in_dimension(0) == [[0., 10.]]).all()
    assert (kicr.cokernel_diagrams().in_dimension(1) == [[30., math.inf]]).all()
    assert (kicr.image_diagrams().in_dimension(0) == [[10., math.inf]]).all()
    assert (len(kicr.image_diagrams().in_dimension(1)) == 0)


if __name__ == "__main__":
    test_kernel_1()
    test_kernel_2()




#!python3
import numpy as np

import oineus as oin
import math

def test_ind_match():
    v0 = oin.Simplex_double([0], 0.2)
    v1 = oin.Simplex_double([1], 0.1)
    v2 = oin.Simplex_double([2], 0.3)

    # indices of vertices should be sorted
    e1 = oin.Simplex_double([0, 1], 0.9)
    e2 = oin.Simplex_double([0, 2], 0.5)
    e3 = oin.Simplex_double([1, 2], 0.8)

    t = oin.Simplex_double([0, 1, 2], 1.0)

    simplices = [v0, v1, v2, e1, e2, e3, t]

    negate = False
    n_threads = 1

    fil_codomain = oin.Filtration_double(simplices, negate, n_threads)

    v0 = oin.Simplex_double([0], 0.5)
    v1 = oin.Simplex_double([1], 0.2)
    v2 = oin.Simplex_double([2], 0.6)

    # indices of vertices should be sorted
    e1 = oin.Simplex_double([0, 1], 0.9)
    e2 = oin.Simplex_double([0, 2], 0.7)
    e3 = oin.Simplex_double([1, 2], 0.91)

    t = oin.Simplex_double([0, 1, 2], 1.2)

    simplices = [v0, v1, v2, e1, e2, e3, t]
    fil_domain = oin.Filtration_double(simplices, negate, n_threads)

    m = oin.get_induced_matching(fil_domain, fil_codomain)

    assert(len(m[0]) == 3)
    assert(len(m[1]) == 1)

if __name__ == "__main__":
    test_ind_match()



from icecream import ic
import time
import matplotlib.pyplot as plt
import oineus as oin
import numpy as np
import torch
import jax
import jax.numpy as jnp
import jax.random as jrandom
from jax import vmap,jit
import equinox as eqx
from ripser import ripser
from persim import plot_diagrams
WASS_DIST_Q = 2
DIM = 1
LATENT = 2
REG_PARAM = 1.0
STEPS = 0


#Plotting
def plot_critical_set(pts,step, edges, longest_edges, lengths , crit_values, mask_by_index,ax=plt,diagram=None):
    '''
    Parameters
    ----------
    pts : original data
    fil : filtration
    indices : indices of critical set

    Returns
    -------
    ax : plot
    '''

    if len(edges) == 0:
        return

    #Set up gradients
    l = lengths.detach().numpy()
    val = crit_values.numpy()
    edge_grads = {} 
    grad_values = []
    for key in mask_by_index:
        mask = mask_by_index[key]
        #value = len(l[mask] - val[mask])
        # value = np.sum(np.sign(l[mask]-val[mask]))
        value = np.sum(2*(l[mask] - val[mask]))
        edge_grads[key] = value
        grad_values.append(value)
    max_value = float(np.max(np.abs(grad_values)))
    #Normalize grads
    for key in edge_grads:
        edge_grads[key] = edge_grads[key]/max_value

    #Plotting
    fig = plt.figure(figsize=(15,5))
    fig.suptitle(f'Strategy={directory}, Loss={LOSS_TYPE}, Step={LEARNING_RATE}', fontsize=16)
    dig_axes = fig.add_subplot(1,2,2)
    plot_diagrams([diagram.in_dimension(0),diagram.in_dimension(1)],ax=dig_axes)
   
    pts_axes = fig.add_subplot(1,2,1)
    pts_axes.scatter(pts[:,0],pts[:,1],color='green')
    
    for key in edge_grads:
        edge_color = cm.viridis(edge_grads[key])
        edge = longest_edges[key]
        pts_axes.plot(pts[edge,0],pts[edge,1],c=edge_color,)

    #create colorbar 
    sm = cm.ScalarMappable(cmap=cm.viridis)
    sm.set_array(grad_values)
    # cbar = ax[0].colorbar(sm)
    #plt.plot(pts[:,0][edges.T],pts[:,1][edges.T],color='purple')

    #save figure
    plt.subplots_adjust(hspace=0.5, top=0.9, bottom=0.1)
    
    fig.savefig(f"Crit_sets/{str((step//100)%10) + str((step//10)%10) + str(step % 10) }.jpg")
    pts_axes.set_xlim(left=-1.4,right=1.2)
    fig.clf()
    plt.close(fig)
    return 
    
def plot_gradients(X,grad,step):
    grad = -LEARNING_RATE*grad
    plt.quiver(X[:,0],X[:,1], grad[:,0], grad[:,1],scale_units='xy',scale=0.01)
    plt.savefig(f'Gradients/Grad_at_{step}_{time.time()}.jpg')
    plt.clf()
    return

def diagram_plot(diagram,show=True):
    plt.clf()
    diagram_np = [diagram.in_dimension(0),diagram.in_dimension(1)]
    plot_diagrams(diagram_np)
    if show:
        plt.show()
    
def diag_grad(diagram,index_diagram,indices,values):
    #Draw arrows on the persistence diagram indicating the gradient direction of a point on the diagram
    indices_to_values = { indices[i]: values[i] for i in range(len(indices))}
    for (i,pair) in enumerate(index_diagram):
        point = diagram[i]
        # ic(point)
        birth_simplex, death_simplex = pair
        try:
            new_point = np.asarray([indices_to_values[birth_simplex],indices_to_values[death_simplex]])
        except:
            new_point = point
        grad = new_point - point
        plt.scatter(point[0],point[1],color='blue') 
        plt.plot([point[0],new_point[0]],[point[1],new_point[1]],color='black')
        plt.scatter(new_point[0],new_point[1],color='red')

def out_grid(model):
    Z = np.linspace(0,10.0,num=40)
    B = np.meshgrid(Z,Z)
    XX,YY = B[0].flatten(), B[1].flatten()
    grid = np.vstack([XX,YY]).T
    output = [ vec.detach().numpy() for vec in list(map(model.decode,torch.Tensor(grid).detach()))]
    return np.asarray(output) 
def save_outputs(outputs):
    plt.clf()
    for (step,output) in enumerate(outputs):
        out = output.detach().numpy()
        plt.scatter(out[:,0],out[:,1])
        plt.savefig(f'outputs/{str((step//100)%10) + str((step//10)%10) + str(step % 10) }.jpg')
        plt.clf()
    
def save_latent(latent_coords):
    global STEPS
    step = STEPS
    plt.clf()
    plt.scatter(latent_coords[:,0],latent_coords[:,1])
    plt.savefig(f"outputs/{str((step//100)%10) + str((step//10)%10) + str(step % 10) }.jpg")
    plt.clf()
    
#Topology Utils
def create_mask_by_index(og_crit_indices):
    crit_values_by_index = {}  
    for (i,index) in enumerate(og_crit_indices):
        if index not in crit_values_by_index:
            crit_values_by_index[index]  = []
        crit_values_by_index[index].append(i)
    return crit_values_by_index


#Model
class AutoEnc(torch.nn.Module):
    def __init__(self, in_size):
        super().__init__()
        self.encode = torch.nn.Sequential(torch.nn.Linear(in_size,LATENT),
                        torch.nn.ReLU(),
                        torch.nn.Linear(LATENT,LATENT)
                     )

        self.decode = torch.nn.Sequential(torch.nn.Linear(LATENT,in_size), 
                        torch.nn.ReLU(),
                        torch.nn.Linear(in_size,in_size)
                      )
    # def encode(self,x):
        # for layer in self.enc_layers:
            # x = layer(x)
        # return x
    # def decode(self,z):
        # for layer in self.dec_layers:
            # z = layer(z)
        # return z
    def __call__(self,x):
        #Returns the reconstruction of x
        x = self.encode(x)
        x = self.decode(x)
        return x
        

#Evalutation
def topological_loss(pts, space_diagram, model):
    global STEPS
    latent_coords = model.encode(pts)
    save_latent(latent_coords.detach().numpy())
    temp_latent_coords = np.asarray(latent_coords.detach().numpy(),dtype=np.double) 
    fil,longest_edges = oin.get_vr_filtration_and_critical_edges(temp_latent_coords, max_dim=2, max_radius=np.inf, n_threads=4)
    top_opt = oin.TopologyOptimizer(fil)
    latent_diagram = top_opt.compute_diagram(include_inf_points=False)
    latent_index_diagram = top_opt.compute_index_diagram(include_inf_points=False,include_zero_persistence_points=False)

    ic(latent_index_diagram[DIM])
    maximum_scale = np.max(latent_diagram[DIM])
    space_diagram = maximum_scale*space_diagram
    diag_list = []
    
    for i in range(space_diagram.shape[0]):
        birth =space_diagram[i][0]
        death =space_diagram[i][1]
        diag_list.append(oin.DiagramPoint_double(birth,death))

    indices, values = top_opt.match(diag_list, DIM, WASS_DIST_Q )

    #
    plt.clf()
    plt.scatter(space_diagram[:,0],space_diagram[:,1],color='orange')
    diag_grad(latent_diagram[DIM],latent_index_diagram[DIM],indices,values)
    plt.savefig(f"diags/{str((STEPS//100)%10) + str((STEPS//10)%10) + str(STEPS % 10) }.jpg")
    plt.clf()
    #
    
    critical_sets = top_opt.singletons(indices,values)
    crit_indices, crit_values = top_opt.combine_loss(critical_sets, oin.ConflictStrategy.Max)

    #Sorted index values
    og_crit_indices = np.array(crit_indices, dtype=np.int32)  
    mask_by_index = create_mask_by_index(og_crit_indices)
    crit_edges = longest_edges[og_crit_indices, :]
    crit_edges_s, crit_edges_t = crit_edges[:, 0], crit_edges[:, 1]
    crit_values = torch.Tensor(crit_values)
    lengths = torch.sum((latent_coords[crit_edges_s, :] - latent_coords[crit_edges_t, :])**2, axis=1) ** 0.5
    #plot_critical_set(X,0, crit_edges, longest_edges, lengths, crit_values, mask_by_index, diagram=dgm)
    #top_loss = torch.mean(torch.abs(lengths - crit_values))
    top_loss = torch.mean((lengths - crit_values)**2)
    STEPS = STEPS + 1
    return top_loss

def manual_topological_loss(pts, space_diagram, model):
    global STEPS
    latent_coords = model.encode(pts)
    save_latent(latent_coords.detach().numpy())
    temp_latent_coords = np.asarray(latent_coords.detach().numpy(),dtype=np.double) 
    fil,longest_edges = oin.get_vr_filtration_and_critical_edges(temp_latent_coords, max_dim=2, max_radius=np.inf, n_threads=4)
    top_opt = oin.TopologyOptimizer(fil)
    latent_diagram = top_opt.compute_diagram(include_inf_points=False)
    latent_index_diagram = top_opt.compute_index_diagram(include_inf_points=False,include_zero_persistence_points=False)

    maximum_scale = np.max(latent_diagram[DIM])
    space_diagram = maximum_scale*space_diagram
    lifetimes = np.asarray([death-birth for (birth,death) in space_diagram])
    feature_point = space_diagram[np.argmax(lifetimes)]

    latent_diagram_in_DIM = latent_diagram[DIM]
    latent_lifetimes = np.asarray([death-birth for (birth,death) in latent_diagram_in_DIM])
    most_pers_pair = latent_index_diagram[DIM][np.argmax(latent_lifetimes)] 
    indices, values = most_pers_pair, feature_point
    #
    plt.clf()
    plt.scatter(space_diagram[:,0],space_diagram[:,1],color='orange')
    diag_grad(latent_diagram[DIM],latent_index_diagram[DIM],indices,values)
    plt.savefig(f"diags/{str((STEPS//100)%10) + str((STEPS//10)%10) + str(STEPS % 10) }.jpg")
    plt.clf()
    #
    
    critical_sets = top_opt.singletons(indices,values)
    crit_indices, crit_values = top_opt.combine_loss(critical_sets, oin.ConflictStrategy.Max)

    #Sorted index values
    og_crit_indices = np.array(crit_indices, dtype=np.int32)  
    mask_by_index = create_mask_by_index(og_crit_indices)
    crit_edges = longest_edges[og_crit_indices, :]
    crit_edges_s, crit_edges_t = crit_edges[:, 0], crit_edges[:, 1]
    crit_values = torch.Tensor(crit_values)
    lengths = torch.sum((latent_coords[crit_edges_s, :] - latent_coords[crit_edges_t, :])**2, axis=1) ** 0.5
    #plot_critical_set(X,0, crit_edges, longest_edges, lengths, crit_values, mask_by_index, diagram=dgm)
    #top_loss = torch.mean(torch.abs(lengths - crit_values))
    top_loss = torch.mean((lengths - crit_values)**2)
    STEPS = STEPS + 1
    return top_loss



def reconstruction_loss(pts,reconstruction):
    MSE = torch.nn.MSELoss()
    return MSE(pts,reconstruction)
     
def regularizer(pts,model):
    return

#Training
def pre_train(pts,model):
    epochs = 100
    outputs = []
    losses = []
    optimizer = torch.optim.Adam(model.parameters(),
                                 lr = 1e-1,
                                 weight_decay = 1e-8)
    for epoch in range(epochs):
        optimizer.zero_grad()
        reconstruction = model(pts)
        loss = reconstruction_loss(pts,reconstruction)
        loss.backward()
        optimizer.step()
        losses.append(loss.detach().numpy())
        outputs.append(reconstruction)
    return losses, outputs, model

def train(pts,model,losses,outputs):
    epochs = 30
    optimizer = torch.optim.Adam(model.parameters(),
                                 lr = 1e-1,
                                 weight_decay = 1e-8)
    #computing diagram for points

    X = np.asarray(pts.detach().numpy(), dtype=np.double)
    fil = oin.get_vr_filtration(X,max_dim=2,max_radius=np.inf,n_threads=4)
    top_opt = oin.TopologyOptimizer(fil)
    og_diags = top_opt.compute_diagram(include_inf_points=False)
    # diagram_plot(og_diags,show=False)
    # plt.savefig("og_diagram.jpg")
    # plt.clf()
    
    og_diag = og_diags[DIM]
    og_maximum_scale = np.max(np.max(og_diag))
    og_diag = og_diag/og_maximum_scale
    plt.clf()
    plt.scatter(og_diag[:,0],og_diag[:,1],color='orange')
    plt.plot([0,1.5],[0,1.5],color='black')
    plt.xlim([0.0,1.5])
    plt.ylim([0.0,1.5])
    plt.savefig('og_diagram.jpg')
    plt.clf()
    global STEPS
    STEPS = 0
    for epoch in range(epochs):
        optimizer.zero_grad()
        reconstruction = model(pts)
        loss = reconstruction_loss(pts,reconstruction) + REG_PARAM*manual_topological_loss(pts,og_diag,model)
        loss.backward()
        optimizer.step()
        print(f"Grads applied {STEPS}")
        losses.append(loss.detach().numpy())
        outputs.append(reconstruction)
    return losses, outputs, model
 



if __name__=="__main__2":
    ic(STEPS)
    theta = np.random.random((100,))*2*np.pi
    Mat = np.random.random((2,3))
    pts = np.vstack([np.cos(theta),np.sin(theta)]).T
    pts = pts @ Mat
    pts = torch.Tensor(pts)
    max_dim = 2
    empty = oin.Diagrams_double(DIM+1)[DIM]
    aut = AutoEnc(pts.shape[1])
    losses, outputs, aut = pre_train(pts,aut)
    out = aut.encode(pts).detach().numpy()
    plt.scatter(out[:,0],out[:,1])
    plt.savefig(f"pretrained.jpg")
    plt.show()
    losses, outputs, aut = train(pts,aut,losses,outputs)
    plt.plot(losses)
    plt.show()

if __name__=="__main__":
    torch.manual_seed(45)
    np.random.seed(45)
    dat = np.loadtxt("example_data/annulus/dat.txt")[0:50]
    mat = np.random.random((2,3))
    pts = dat @ mat
    pts = torch.Tensor(pts)
    aut = AutoEnc(pts.shape[1])
    losses, outputs, aut = pre_train(pts,aut)
    out = aut.encode(pts).detach().numpy()
    plt.scatter(out[:,0],out[:,1])
    plt.savefig(f"pretrained.jpg")
    plt.clf()
    losses, outputs, aut = train(pts,aut,losses,outputs)
    plt.plot(losses)
    plt.show()
    
        



#!/usr/bin/env python3

import numpy as np
import oineus as oin

def dion_dgm_to_numpy(dion_dgm):
    return np.array([[p.birth, p.death] for p in dion_dgm ]).astype(np.float32)

def prepare_data():
    import dionysus as dion
    top_dim = 3
    for negate in [True, False]:
        a = np.load("vertebra_32x32x32_float32.npy")
        # compute diagrams with Dionysis
        fil_us = dion.fill_freudenthal(a, reverse=negate)
        p = dion.homology_persistence(fil_us)
        dion_dgms = dion.init_diagrams(p, fil_us)
        for dim in range(top_dim):
            dgm = dion_dgm_to_numpy(dion_dgms[dim])
            if negate:
                dgm[dgm == np.inf] = -np.inf
            if negate:
                np.save(f"dgm_vertebra_32_neg_{dim}.npy", dgm)
            else:
                np.save(f"dgm_vertebra_32_{dim}.npy", dgm)


def sort_numpy_array(a: np.ndarray):
    assert a.ndim == 2
    return a[np.lexsort((a[:, 1], a[:, 0]))]


def test_vertebra():
    top_dim = 3
    wrap = False
    a = np.load("vertebra_32x32x32_float32.npy")
    for dualize in [True, False]:
        for n_threads in [1, 7]:
            for negate in [False, True]:
                for clearing_opt in [True]:
                    print(f"{dualize=}, {negate=}, {n_threads=}, {clearing_opt=}")
                    rp = oin.ReductionParams()
                    rp.n_threads = n_threads
                    rp.clearing_opt = clearing_opt
                    oin_dgms = oin.compute_diagrams_ls(a, negate, wrap, top_dim-1, rp, include_inf_points=True, dualize=dualize)
                    for dim in range(top_dim):
                        if negate:
                            dion_dgm_fname = f"dgm_vertebra_32_neg_{dim}.npy"
                        else:
                            dion_dgm_fname = f"dgm_vertebra_32_{dim}.npy"
                        dion_dgm = np.load(dion_dgm_fname)
                        oin_dgm = oin_dgms[dim]

                        dion_dgm = sort_numpy_array(dion_dgm)
                        oin_dgm = sort_numpy_array(oin_dgm)
                        assert dion_dgm.shape == oin_dgm.shape

                        dion_dgm[dion_dgm == np.inf] = 0.0
                        dion_dgm[dion_dgm == -np.inf] = 0.0
                        oin_dgm[oin_dgm == np.inf] = 0.0
                        oin_dgm[oin_dgm == -np.inf] = 0.0

                        assert np.linalg.norm(dion_dgm - oin_dgm) < 0.00001



if __name__ == "__main__":
    test_vertebra()


from icecream import ic
import oineus as oin
import numpy as np
import matplotlib.pyplot as plt
DIM = 1
WASS_DIST_Q = 2
max_dim = 1
empty = oin.Diagrams_double(DIM+1)[DIM]
Z = np.random.random((100,2))
fil,longest_edges = oin.get_vr_filtration_and_critical_edges(Z, max_dim=2, max_radius=np.inf, n_threads=4)
top_opt = oin.TopologyOptimizer(fil)
diagram = top_opt.compute_diagram(include_inf_points=False)
#Same loss as before with edges, i.e. align edges to their critical lengths
indices, values = top_opt.match(empty, DIM, WASS_DIST_Q )
ic(indices,values)




import numpy as np
import plotly.graph_objs as go
import dash
from dash import html, dcc
from dash.dependencies import Input, Output, State

import oineus as oin


# Number of input points
num_input_points = 40
avg_r = 1.0
low_r = avg_r - 0.2
high_r = avg_r + 0.2
# Generate input points sampled from a circle
theta = np.linspace(0, 2*np.pi, num_input_points)
r = np.random.uniform(low=low_r, high=high_r, size=num_input_points)  # Uniform distribution within the circle
# r = avg_r
input_points = np.column_stack((r * np.cos(theta), r * np.sin(theta)))

print(f"{input_points.shape = }")

max_dim = 2
max_radius = 2.0 * high_r
n_threads = 1

init_dim = 1

fil, crit_edges = oin.get_vr_filtration_and_critical_edges(input_points, max_dim, max_radius, n_threads)

print(f"{fil.size() = }")

top_opt = oin.TopologyOptimizer(fil)
top_opt.reduce_all()
dgms = top_opt.compute_diagram(False)

diagram_points = dgms[init_dim]

# Rule to map a diagram point to a set of edges between input points
def get_edges_crit(x, y, change_coord, dim):
    for pt in dgms.in_dimension(dim, False):
        if np.abs(pt.birth - x) + np.abs(pt.death - y) < 0.0001:
            birth_idx = pt.birth_index
            death_idx = pt.death_index
            break

    if change_coord == "inc-birth":
        simplex_inds = top_opt.increase_birth(birth_idx)
    elif change_coord == "dec-birth":
        simplex_inds = top_opt.decrease_birth(birth_idx)
    elif change_coord == "inc-death":
        simplex_inds = top_opt.increase_death(death_idx)
    elif change_coord == "dec-death":
        simplex_inds = top_opt.decrease_death(death_idx)

    edges = [(input_points[e[0]], input_points[e[1]]) for e in crit_edges[simplex_inds]]
    return edges


def get_edges_dgm(x, y, change_coord, dim):
    for pt in dgms.in_dimension(dim, False):
        if np.abs(pt.birth - x) + np.abs(pt.death - y) < 0.0001:
            birth_idx = pt.birth_index
            death_idx = pt.death_index
            break

    if change_coord in ["inc-birth", "dec-birth"]:
        simplex_inds = [birth_idx]
    elif change_coord in ["inc-death", "dec-death"]:
        simplex_inds = [death_idx]

    edges = [(input_points[e[0]], input_points[e[1]]) for e in crit_edges[simplex_inds]]
    return edges


def get_triangles_crit(x, y, change_coord, dim):
    if dim != 1 or change_coord not in ["inc-death", "dec-death"]:
        raise RuntimeError("wrong arguments")

    for pt in dgms.in_dimension(dim, False):
        if np.abs(pt.birth - x) + np.abs(pt.death - y) < 0.0001:
            birth_idx = pt.birth_index
            death_idx = pt.death_index
            break

    if change_coord == "inc-death":
        simplex_inds = top_opt.increase_death(death_idx)
    elif change_coord == "dec-death":
        simplex_inds = top_opt.decrease_death(death_idx)

    triangles = []
    for t_idx in simplex_inds:
        t_vertices = fil.get_cell(t_idx).vertices
        triangles.append((input_points[t_vertices[0]], input_points[t_vertices[1]], input_points[t_vertices[2]]))

    return triangles


def get_triangles_dgm(x, y, change_coord, dim):
    if dim != 1 or change_coord not in ["inc-death", "dec-death"]:
        raise RuntimeError("wrong arguments")

    for pt in dgms.in_dimension(dim, False):
        if np.abs(pt.birth - x) + np.abs(pt.death - y) < 0.0001:
            birth_idx = pt.birth_index
            death_idx = pt.death_index
            break

    triangles = []

    for t_idx in [death_idx]:
        t_vertices = fil.get_cell(t_idx).vertices
        triangles.append((input_points[t_vertices[0]], input_points[t_vertices[1]], input_points[t_vertices[2]]))

    return triangles


app = dash.Dash(__name__)

app.layout = html.Div([
    html.Div([
        html.Div([
            html.Label('Change direction:'),
            dcc.RadioItems(
                id='change-coord',
                options=[
                    {'label': 'Increase Birth', 'value': 'inc-birth'},
                    {'label': 'Decrease Birth', 'value': 'dec-birth'},
                    {'label': 'Increase Death', 'value': 'inc-death'},
                    {'label': 'Decrease Death', 'value': 'dec-death'},
                ],
                value='inc-birth',
                labelStyle={'display': 'inline-block'}
            ),
        ], style={'display': 'inline-block', 'margin-right': '20px'}),
        html.Div([
            html.Label('Dimension:'),
            dcc.RadioItems(
                id='dimension',
                options=[
                    {'label': '0', 'value': '0'},
                    {'label': '1', 'value': '1'},
                    # {'label': '2', 'value': '2'}
                ],
                value=str(init_dim),
                labelStyle={'display': 'inline-block'}
            ),
        ], style={'display': 'inline-block', 'margin-right': '20px'}),
        html.Div([
            html.Label('Critical set vs diagram method:'),
            dcc.RadioItems(
                id='critical-set',
                options=[
                    {'label': 'Critical set', 'value': 'crit-set'},
                    {'label': 'Diagram method', 'value': 'dgm-method'},
                ],
                value='crit-set',
                labelStyle={'display': 'inline-block'}
            ),
        ], style={'display': 'inline-block', 'margin-right': '20px'}),
        html.Div([
            html.Label('For death/dim 1: triangles or longest edges:'),
            dcc.RadioItems(
                id='plot-chains',
                options=[
                    {'label': 'Triangles', 'value': 'chains'},
                    {'label': 'Edges', 'value': 'edges'},
                ],
                value='chains',
                labelStyle={'display': 'inline-block'}
            ),
        ], style={'display': 'inline-block'}),
    ], style={'margin-bottom': '20px'}),
    dcc.Graph(
        id='point-cloud',
        figure={
            'data': [go.Scatter(x=input_points[:, 0], y=input_points[:, 1], mode='markers', marker=dict(color='black', size=5))],
            'layout': go.Layout(title='Input Points', xaxis=dict(range=[-2, 2]), yaxis=dict(range=[-2, 2]))
        }
    ),
    dcc.Graph(
        id='diagram-plot',
        figure={
            'data': [go.Scatter(x=diagram_points[:, 0], y=diagram_points[:, 1], mode='markers', marker=dict(color='red', size=10))],
            'layout': go.Layout(title='Diagram Points', xaxis=dict(range=[0, 5]), yaxis=dict(range=[0, 5]))
        }
    ),
])


@app.callback(
    Output('diagram-plot', 'figure'),
    [Input('dimension', 'value')],
    prevent_initial_call=False
)
def update_diagram_points(dimension_value):
    global diagram_points
    # Get the diagram points for the selected dimension
    dim = int(dimension_value)
    diagram_points = dgms[dim]

    # Create a new scatter plot trace for the diagram points
    trace = go.Scatter(
        x=diagram_points[:, 0],
        y=diagram_points[:, 1],
        mode='markers',
        marker=dict(color='red', size=8)
    )
    # Update the figure
    return {
        'data': [trace],
        'layout': go.Layout(title='Diagram Points', xaxis=dict(range=[-0.1, 2 * high_r]), yaxis=dict(range=[-0.1, 2 * high_r]))
    }


@app.callback(
    Output('point-cloud', 'figure'),
    [Input('diagram-plot', 'hoverData'),
     Input('change-coord', 'value'),
     Input('dimension', 'value'),
     Input('critical-set', 'value'),
     Input('plot-chains', 'value')],
    [State('point-cloud', 'figure')],
    prevent_initial_call=True
)
def display_edges(hoverData, change_coord, dimension, crit_or_dgm, plot_chains, existing_figure):
    global dgms
    # Copy the existing figure
    figure = dict(existing_figure)

    # Remove any existing edge traces
    figure['data'] = [trace for trace in figure['data'] if trace['type'] != 'scatter' or 'line' not in trace['mode']]

    dim = int(dimension)

    if hoverData:
        point_index = hoverData['points'][0]['pointIndex']
        x, y = diagram_points[point_index]


        # for 2D data we only need triangles if we modify death value in 1D diagram
        # in all other cases we draw longest edges
        if dim == 1 and change_coord in ["dec-death", "inc-death"] and plot_chains == "chains":
            if crit_or_dgm == "crit-set":
                triangles = get_triangles_crit(x, y, change_coord, dim)
            elif crit_or_dgm == "dgm-method":
                triangles = get_triangles_dgm(x, y, change_coord, dim)

            # Create line segments for the triangles
            line_x = []
            line_y = []
            for triangle in triangles:
                for i in range(3):  # 3 vertices in a triangle
                    line_x.extend([triangle[i][0], triangle[(i+1)%3][0], None])  # %3 to ensure we cycle back to the first vertex
                    line_y.extend([triangle[i][1], triangle[(i+1)%3][1], None])

            # Add triangles to the input plot
            figure['data'].append(go.Scatter(x=line_x, y=line_y, mode='lines', line=dict(color='blue')))

        else:
            if crit_or_dgm == "crit-set":
                edges = get_edges_crit(x, y, change_coord, dim)
            elif crit_or_dgm == "dgm-method":
                edges = get_edges_dgm(x, y, change_coord, dim)

            for edge in edges:
                edge_x, edge_y = zip(*edge)
                edge_trace = go.Scatter(x=edge_x, y=edge_y, mode='lines', line=dict(color='blue', width=2))
                figure['data'].append(edge_trace)

    return figure

if __name__ == '__main__':
    app.run_server(debug=True)




import functools as ft

import numpy as np
import eagerpy as epy

from .. import _oineus
from .. import vr_filtration as non_diff_vr_filtration

# to copy docstring and name from the wrapped _oineus (C++) methods


class DiffFiltration:
    def __init__(self, fil, values):
        self.under_fil = fil
        self.values = values

    def __len__(self):
        return len(self.under_fil)

    def __repr__(self):
        return f"DiffFil(under_fil={self.under_fil}, values={self.values})"

    def __iter__(self):
        return iter(self.under_fil)

    #@ft.wraps(_oineus.Filtration.max_dim)
    def max_dim(self):
        return self.under_fil.max_dim()

    #@ft.wraps(_oineus.Filtration.size)
    def size(self):
        return self.under_fil.size()

    #@ft.wraps(_oineus.Filtration.size_in_dimension)
    def size_in_dimension(self, dim):
        return self.under_fil.size(dim)

    #@ft.wraps(_oineus.Filtration.n_vertices)
    def n_vertices(self):
        return self.under_fil.n_vertices()

    #@ft.wraps(_oineus.Filtration.cells)
    def cells(self):
        return self.under_fil.cells()

    #@ft.wraps(_oineus.Filtration.get_id_by_sorted_id)
    def id_by_sorted_id(self, sorted_id):
        return self.under_fil.id_by_sorted_id(sorted_id)

    #@ft.wraps(_oineus.Filtration.get_sorted_id_by_id)
    def sorted_id_by_id(self, id):
        return self.under_fil.sorted_id_by_id(id)

    #@ft.wraps(_oineus.Filtration.get_cell)
    def cell(self, sorted_idx):
        return self.under_fil.cell(sorted_idx)

    #@ft.wraps(_oineus.Filtration.get_simplex)
    def simplex(self, sorted_idx):
        return self.under_fil.simplex(sorted_idx)

    #@ft.wraps(_oineus.Filtration.get_sorting_permutation)
    def sorting_permutation(self):
        return self.under_fil.sorting_permutation()

    #@ft.wraps(_oineus.Filtration.get_inv_sorting_permutation)
    def get_inv_sorting_permutation(self):
        return self.under_fil.inv_sorting_permutation()

    #@ft.wraps(_oineus.Filtration.cell_by_uid)
    def cell_by_uid(self, uid):
        return self.under_fil.cell_by_uid(uid)

    #@ft.wraps(_oineus.Filtration.boundary_matrix)
    def boundary_matrix(self, uid):
        return self.under_fil.boundary_matrix(uid)

    #@ft.wraps(_oineus.Filtration.simplex_value_by_sorted_id)
    def simplex_value_by_sorted_id(self, sorted_id):
        return self.under_fil.simplex_value_by_sorted_id(sorted_id)

    #@ft.wraps(_oineus.Filtration.simplex_value_by_vertices)
    def simplex_value_by_uid(self, uid):
        return self.under_fil.simplex_value_by_uid(uid)

    def cell_value_by_uid(self, uid):
        return self.under_fil.cell_value_by_uid(uid)

    #@ft.wraps(_oineus.Filtration.get_sorted_id_by_vertices)
    def sorted_id_by_uid(self, uid):
        return self.under_fil.sorted_id_by_uid(uid)

    #@ft.wraps(_oineus.Filtration.reset_ids_to_sorted_ids)
    def reset_ids_to_sorted_ids(self):
        self.under_fil.reset_ids_to_sorted_ids()


class TopologyOptimizer:
    """
    A wrapper around C++ topology optimizer
    The C++ class is templated by the Cell type,
    the purpose of this class is to allow the user
    not to worry about the type of a filtration
    """
    def __init__(self, fil):
        if isinstance(fil, DiffFiltration):
            fil = fil.under_fil
        if type(fil) is _oineus.Filtration:
            self.under_opt = _oineus.TopologyOptimizer(fil)
        elif type(fil) is _oineus.ProdFiltration:
            self.under_opt = _oineus.TopologyOptimizerProd(fil)
        else:
            raise RuntimeError("unknown filtration type in oineus.diff.TopologyOptimizer constructor")

    #@ft.wraps(_oineus.TopologyOptimizer.compute_diagram)
    def compute_diagram(self, include_inf_points: bool):
        return self.under_opt.compute_diagram(include_inf_points)

    #@ft.wraps(_oineus.TopologyOptimizer.simplify)
    def simplify(self, epsilon: float, strategy, dim: int):
        return self.under_opt.simplify(epsilon, strategy, dim)

    #@ft.wraps(_oineus.TopologyOptimizer.get_nth_persistence)
    def get_nth_persistence(self, dim: int, n: int):
        return self.under_opt.get_nth_persistence(dim=dim, n=n)

    #@ft.wraps(_oineus.TopologyOptimizer.match)
    def match(self, template_dgm, dim: int, wasserstein_q: float=1.0, return_wasserstein_distance: bool=False):
        return self.under_opt.match(temlate_dgm=template_dgm, dim=dim, wasserstein_q=wasserstein_q,
                                    return_wasserstein_distance=return_wasserstein_distance)

    @property
    def homology_decomposition(self):
        return self.under_opt.homology_decomposition

    @property
    def cohomology_decomposition(self):
        return self.under_opt.cohomology_decomposition

    #@ft.wraps(_oineus.TopologyOptimizer.singleton)
    def singleton(self, index: int, value: float):
        return self.under_opt.singleton(index, value)

    #@ft.wraps(_oineus.TopologyOptimizer.singletons)
    def singletons(self, indices, values):
        return self.under_opt.singletons(indices, values)

    #@ft.wraps(_oineus.TopologyOptimizer.reduce_all)
    def reduce_all(self):
        return self.under_opt.reduce_all()

    #@ft.wraps(_oineus.TopologyOptimizer.update)
    def update(self):
        return self.under_opt.update()

    #@ft.wraps(_oineus.TopologyOptimizer.increase_death)
    def increase_death(self, negative_simplex_idx: int):
        return self.under_opt.increase_death(negative_simplex_idx)

    #@ft.wraps(_oineus.TopologyOptimizer.decrease_death)
    def decrease_death(self, negative_simplex_idx: int):
        return self.under_opt.decrease_death(negative_simplex_idx)

    #@ft.wraps(_oineus.TopologyOptimizer.increase_birth)
    def increase_birth(self, positive_simplex_idx: int):
        return self.under_opt.increase_birth(positive_simplex_idx)

    #@ft.wraps(_oineus.TopologyOptimizer.decrease_birth)
    def decrease_birth(self, positive_simplex_idx: int):
        return self.under_opt.decrease_birth(positive_simplex_idx)

    #@ft.wraps(_oineus.TopologyOptimizer.combine_loss)
    def combine_loss(self, critical_sets, strategy):
        return self.under_opt.combine_loss(critical_sets, strategy)


def min_filtration(fil_1: DiffFiltration, fil_2: DiffFiltration) -> DiffFiltration:
    fil_1_under = fil_1.under_fil
    fil_2_under = fil_2.under_fil

    min_fil_under, inds_1, inds_2 = _oineus._min_filtration_with_indices(fil_1_under, fil_2_under)

    inds_1 = np.array(inds_1)
    inds_2 = np.array(inds_2)

    vals_1 = epy.astensor(fil_1.values)[inds_1]
    vals_2 = epy.astensor(fil_2.values)[inds_2]

    min_fil_values = epy.min(epy.stack((vals_1, vals_2)), axis=0).raw

    return DiffFiltration(min_fil_under, min_fil_values)


def freudenthal_filtration(data, negate, wrap, max_dim, n_threads):
    data = epy.astensor(data)
    dim_part = data.ndim
    np_data = data.float64().numpy()
    type_part = "double"
    func = getattr(_oineus, f"get_fr_filtration_and_critical_vertices_{type_part}_{dim_part}")
    fil, cv = func(np_data, negate, wrap, max_dim, n_threads)
    cv = np.array(cv, dtype=np.int64)
    values = data[cv].raw
    return DiffFiltration(fil, values)


def vr_filtration(data, from_pwdists: bool=False, max_dim: int=-1, max_diameter: float=-1.0, eps=1e-6, n_threads=8) -> DiffFiltration:

    data = epy.astensor(data)
    data_np = data.float64().numpy()
    assert(data.ndim == 2)

    fil, edges = non_diff_vr_filtration(data=data_np, from_pwdists=from_pwdists, with_critical_edges=True,
                                        max_dim=max_dim, max_diameter=max_diameter, n_threads=n_threads)

    if not from_pwdists:
        sqdists = epy.sum((data[edges[:, 0].flatten()] - data[edges[:, 1].flatten()]) ** 2 + eps, axis=1)
        diff_dists = epy.sqrt(sqdists).raw

        return DiffFiltration(fil, diff_dists)
    else:
        edges = epy.astensor(edges)
        diff_dists = data[edges[:, 0], edges[:, 1]].raw
        return DiffFiltration(fil, diff_dists)


def mapping_cylinder_filtration(fil_domain: DiffFiltration, fil_codomain: DiffFiltration, v_domain, v_codomain) -> DiffFiltration:
    assert(type(fil_domain) is DiffFiltration)
    assert(type(fil_codomain) is DiffFiltration)

    if isinstance(v_domain, _oineus.Simplex):
        v_domain = v_domain.combinatorial_cell()

    if isinstance(v_codomain, _oineus.Simplex):
        v_codomain = v_codomain.combinatorial_cell()

    under_fil_dom = fil_domain.under_fil
    under_fil_cod = fil_codomain.under_fil

    under_cyl_fil, cyl_val_inds = _oineus._mapping_cylinder_with_indices(under_fil_dom, under_fil_cod, v_domain, v_codomain)

    cyl_val_inds = epy.astensor(np.array(cyl_val_inds, dtype=np.int64))

    concat_vals = epy.concatenate((epy.astensor(fil_domain.values), epy.astensor(fil_codomain.values)))

    assert(concat_vals.ndim == 1 and concat_vals.shape[0] == fil_domain.size() + fil_codomain.size())

    cyl_values = concat_vals[cyl_val_inds].raw

    return DiffFiltration(under_cyl_fil, cyl_values)

